;
;                    Inertia Realtime 3D Rendering Engine
;    Copyright (c) 1996, Alex Chalfin, Jeroen Bouwens. All Rights Reserved.
;          DISTRIBUTION OF THIS SOURCE CODE IS STRICTLY PROHIBITED
;

;
; PERSPASM.INC
; code: Alex Chalfin
;

; Perspective based polygon include file

; paramters:
; edi  - Polygon pointer value

  Mov  eax,[edi].PolyAPtr       ; VA = P->A
  Mov  ebx,[edi].PolyBPtr       ; VB = P->B
  Mov  ecx,[edi].PolyCPtr       ; VC = P->C
                                ;
                                ;
IFDEF  __INTERP_UV__            ;
  Lea  esi,[edi].UV1x           ;  TVA = &P->UV1
  Mov  [_TVA],esi               ;  Store the address of first texture coord
                                ;
  Lea  esi,[edi].UV2x           ;  TVB = &P->UV2
  Mov  [_TVB],esi               ;  Store the address of second texture coord
                                ;
  Lea  esi,[edi].UV3x           ;  TVC = &P->UV3
  Mov  [_TVC],esi               ;  Store the address of third texture coord
ENDIF                           ;
                                ;
;---------------------------------------------------------------------------
; First bubble sort compare
;---------------------------------------------------------------------------
                                ;
  fld  dword ptr [eax].ScrY     ;
  fld  dword ptr [ebx].ScrY     ;
  fcompp                        ; if (VA->Scr.y > VB->Scr.y)
                                ;
  Push  eax                     ; save first vertex pointer
  fstsw ax                      ;
  sahf                          ; store FPU flags
  Pop   eax                     ;
                                ;
  ja  @@SkipPointerSwap1        ; if st(0) <= st(1) then jump
                                ;
  Xchg eax,ebx                  ; Swap the A and B pointers
                                ;
IFDEF __INTERP_UV__             ;
  Mov  edx,[_TVA]               ;
  Xchg edx,[_TVB]               ;
  Mov  [_TVA],edx               ;
ENDIF                           ;
                                ;
 @@SkipPointerSwap1:            ;
                                ;
;---------------------------------------------------------------------------
; Second bubble sort compare
;---------------------------------------------------------------------------
                                ;
  fld  dword ptr [ebx].ScrY     ;
  fld  dword ptr [ecx].ScrY     ;
  fcompp                        ; if (VB->Scr.y > VC->Scr.y)
                                ;
  Push  eax                     ; Save first vertex pointer
  fstsw ax                      ;
  sahf                          ; store FPU flags
  Pop   eax                     ;
                                ;
  ja  @@SkipPointerSwap2        ; if st(0) <= st(1) then jump
                                ;
  Xchg ebx,ecx                  ; Swap the B and C pointers
                                ;
IFDEF __INTERP_UV__             ;
  Mov  edx,[_TVB]               ;
  Xchg edx,[_TVC]               ;
  Mov  [_TVB],edx               ;
ENDIF                           ;
                                ;
 @@SkipPointerSwap2:            ;
                                ;
;---------------------------------------------------------------------------
; Third bubble sort compare
;---------------------------------------------------------------------------
                                ;
  fld  dword ptr [eax].ScrY     ;
  fld  dword ptr [ebx].ScrY     ;
  fcompp                        ; if (VA->Scr.y > VB->Scr.y)
                                ;
  Push  eax                     ; Save first vertex pointer
  fstsw ax                      ;
  sahf                          ; store FPU flags
  Pop   eax                     ;
                                ;
  ja  @@SkipPointerSwap3        ; if st(0) <= st(1) then jump
                                ;
  Xchg eax,ebx                  ; Swap the A and B pointers
                                ;
IFDEF __INTERP_UV__             ;
  Mov  edx,[_TVA]               ;
  Xchg edx,[_TVB]               ;
  Mov  [_TVA],edx               ;
ENDIF                           ;
                                ;
 @@SkipPointerSwap3:            ;
                                ;
;----------------------------------------------------------------------------
; Store the vertex pointers
;----------------------------------------------------------------------------
                                ;
  Mov  [_VertPtrA],eax          ;  Store the vertex pointers into
  Mov  [_VertPtrB],ebx          ;  variables for later use
  Mov  [_VertPtrC],ecx          ;
                                ;
;----------------------------------------------------------------------------
; Calculate gouraud intensities if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF __INTERP_C__              ;
                                ;
  fld   dword ptr [eax].VShade  ; load the first vertex shade intensity
  fmul  dword ptr [__tmp_flt_63] ; scale it up to yield 64 intensities
                                ;
  fld   dword ptr [ebx].VShade  ; second vertex shade value
  fmul  dword ptr [__tmp_flt_63];
                                ;
  fld   dword ptr [ecx].VShade  ; third vertex shade value
  fmul  dword ptr [__tmp_flt_63];
                                ;
  fstp  dword ptr [_FLTInten3]  ; save the values
  fstp  dword ptr [_FLTInten2]  ;
  fstp  dword ptr [_FLTInten1]  ;
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate texture coords if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF __INTERP_UV__             ;
                                ;
  Push eax                      ; save the vertex pointers for later
  Push ebx                      ;
  Push ecx                      ;
  Push edi                      ;
                                ;
  Mov  edx,eax                  ; store vert A pointer
  Mov  edi,ebx                  ; store vert B pointer
  Mov  esi,ecx                  ; store vert C pointer
                                ;
  Mov  eax,[_TVA]               ; grab the texture pointers
  Mov  ebx,[_TVB]               ;
  Mov  ecx,[_TVC]               ;
                                ;
  fld  dword ptr [eax].X2D      ; get the first texture X value
  fmul dword ptr [__tmp_flt_255]; scale it up
  fmul dword ptr [edx].OneOverZ ; multiply by one over z
                                ;
  fld  dword ptr [eax].Y2D      ; get the first texture Y value
  fmul dword ptr [__tmp_flt_255]; scale it up
  fmul dword ptr [edx].OneOverZ ; multiply by one over z
                                ;
                                ;
  fstp dword ptr [_FLTTextureY1] ;  store it in the proper variables
  fstp dword ptr [_FLTTextureX1] ;
                                ;
  fld  dword ptr [ebx].X2D      ;  get the second texture X value
  fmul dword ptr [__tmp_flt_255] ;  scale it up
  fmul dword ptr [edi].OneOverZ ; multiply by one over z
                                ;
  fld  dword ptr [ebx].Y2D      ;  get the second texture Y value
  fmul dword ptr [__tmp_flt_255] ;  scale it up
  fmul dword ptr [edi].OneOverZ ; multiply by one over z
                                ;
  fstp dword ptr [_FLTTextureY2] ;  store it in the proper variables
  fstp dword ptr [_FLTTextureX2] ;
                                ;
  fld  dword ptr [ecx].X2D      ;  get the third texture X value
  fmul dword ptr [__tmp_flt_255] ;  scale it up
  fmul dword ptr [esi].OneOverZ ; multiply by one over z
                                ;
  fld  dword ptr [ecx].Y2D      ;  get the third texture Y value
  fmul dword ptr [__tmp_flt_255] ;  scale it up
  fmul dword ptr [esi].OneOverZ ; multiply by one over z
                                ;
  fstp dword ptr [_FLTTextureY3] ;  store it in the proper variables
  fstp dword ptr [_FLTTextureX3] ;
                                ;
  Pop  edi                      ;
  Pop  ecx                      ; restore the vertex pointer
  Pop  ebx                      ;
  Pop  eax                      ;
                                ;
;----------------------------------------------------------------------------
; create one over Z variables
;----------------------------------------------------------------------------
                                ;
  Mov  esi,[eax].OneOverZ       ;
  Mov  edx,[ebx].OneOverZ       ;
                                ;
  Mov  [_FLTZ1],esi              ; Store Z1
  Mov  [_FLTZ2],edx              ; Store Z2
                                ;
  Mov  edx,[ecx].OneOverZ       ;
  Mov  [_FLTZ3],edx              ; Store Z3
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate environment mapping  coords if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF __INTERP_ENV__            ;
                                ;
  fld  dword ptr [eax].EnvX     ;  get the first env X value
  fmul dword ptr [__tmp_flt_255] ;  scale it up
  fld  dword ptr [eax].EnvY     ;  get the first env Y value
  fmul dword ptr [__tmp_flt_255] ;  scale it up
                                ;
  fstp dword ptr [_FLTEnvY1]     ;  store it in the proper variables
  fstp dword ptr [_FLTEnvX1]     ;
                                ;
  fld  dword ptr [ebx].EnvX     ;  get the second env X value
  fmul dword ptr [__tmp_flt_255] ;  scale it up
  fld  dword ptr [ebx].EnvY     ;  get the second env Y value
  fmul dword ptr [__tmp_flt_255] ;  scale it up
                                ;
  fstp dword ptr [_FLTEnvY2]     ;  store it in the proper variables
  fstp dword ptr [_FLTEnvX2]     ;
                                ;
  fld  dword ptr [ecx].EnvX     ;  get the third Env X value
  fmul dword ptr [__tmp_flt_255] ;  scale it up
  fld  dword ptr [ecx].EnvY     ;  get the third Env Y value
  fmul dword ptr [__tmp_flt_255] ;  scale it up
                                ;
  fstp dword ptr [_FLTEnvY3]     ;  store it in the proper variables
  fstp dword ptr [_FLTEnvX3]     ;
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate height parameters for the polygon
;----------------------------------------------------------------------------
                                ;
  fld  dword ptr [eax].ScrY     ;
  fadd dword ptr [__tmp_half]    ;
  fistp dword ptr [_CeilY1]      ; ceilY1 = FLT2INT(ceil(VA->Scr.y));
                                ;
  fld  dword ptr [ebx].ScrY     ;
  fadd dword ptr [__tmp_half]    ;
  fistp dword ptr [_CeilY2]      ; ceilY2 = FLT2INT(ceil(VB->Scr.y));
                                ;
  fld  dword ptr [ecx].ScrY     ;
  fadd dword ptr [__tmp_half]    ;
  fistp dword ptr [_CeilY3]      ; ceilY3 = FLT2INT(ceil(VC->Scr.y));
                                ;
  Mov edx,[_CeilY3]              ;
  Mov esi,[_CeilY2]              ;
                                ;
  Sub edx,[_CeilY1]              ; Calculate the Y size of the polygon
  Sub esi,[_CeilY1]              ; calc the Y size of the top
                                ;
  Mov [_TotalHeight],edx         ;
  Mov [_TopHeight],esi           ;
                                ;
  Sub edx,esi                   ;
  Mov [_BottomHeight],edx        ; calc the Y size of the bottom
                                ;
  Mov edx,[_TotalHeight]         ;
  Cmp edx,0                     ;
  Jle @@ExitPolygonFiller       ;
                                ;
;----------------------------------------------------------------------------
; Calculate the delta for the long side of the polygon
;----------------------------------------------------------------------------
                                ;
  fld  dword ptr [ecx].ScrY     ;
  fsub dword ptr [eax].ScrY     ;
                                ;
  fld1                          ;
  fdivrp st(1),st(0)            ; OneOverTotalDY  = 1.0 / (VC->Scr.y - VA->Scr.y);
                                ;
  fst    [_OneOverTotalDY]      ;
                                ;
  fld    dword ptr [ecx].ScrX   ;
  fsub   dword ptr [eax].ScrX   ;
  fmulp  st(1),st(0)            ; (VC->Scr.x - VA->Scr.x) * OneOverDY;
                                ;
  fst    dword ptr [_FLTLongDX]  ; save floating value for sub-pixel correction
                                ;
  fld    st(0)                  ;
  fmul   dword ptr [__tmp_fixone];
  fistp  dword ptr [_FXDLongDX]  ;  convert the delta value to fixed
                                ;
  fild   dword ptr [_CeilY1]     ;
  fsub   dword ptr [eax].ScrY   ;
  fmulp  st(1),st(0)            ; (ceilY1 - VA->Scr.y) * t
  fadd   dword ptr [eax].ScrX   ;
  fmul   dword ptr [__tmp_fixone];
                                ;
  fistp  dword ptr [_FXDLongX]   ; longx  = FLT2FXD(VA->Scr.x + (ceilY1 - VA->Scr.y) * t);
                                ;
;----------------------------------------------------------------------------
; Calculate x delta value for the polygon
;----------------------------------------------------------------------------
                                ;
                                ; calculate 2d cross product to determine
                                ; orientation of polygon
  fld    dword ptr [ecx].ScrX   ;
  fsub   dword ptr [eax].ScrX   ; (VC->Scr.x - VA->Scr.x)
  fld    dword ptr [ebx].ScrY   ;
  fsub   dword ptr [eax].ScrY   ; (VB->Scr.y - VA->Scr.y)
  fmulp  st(1),st(0)            ;
                                ;
  fld    dword ptr [ecx].ScrY   ;
  fsub   dword ptr [eax].ScrY   ; (VC->Scr.y - VA->Scr.y)
  fld    dword ptr [ebx].ScrX   ;
  fsub   dword ptr [eax].ScrX   ; (VB->Scr.x - VA->Scr.x)
  fmulp  st(1),st(0)            ;
                                ;
  fsubrp st(1),st(0)            ;
                                ;
  fld1                          ;
  fdivrp st(1),st(0)            ;
                                ;
  fstp   dword ptr [_PolyDX]     ;
                                ;
;----------------------------------------------------------------------------
; Calculate horizontal gouraud delta value for the polygon if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF __INTERP_C__              ;
                                ;
  fld   dword ptr [_FLTInten3]   ;
  fsub  dword ptr [_FLTInten1]   ; (c3 - c1)
  fld   dword ptr [ebx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VB->Scr.y - VA->Scr.y)
  fmulp st(1),st(0)             ;
                                ;
  fld   dword ptr [ecx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VC->Scr.y - VA->Scr.y)
  fld   dword ptr [_FLTInten2]   ;
  fsub  dword ptr [_FLTInten1]   ; (c2 - c1)
  fmulp st(1),st(0)             ;
                                ;
  fsubrp st(1),st(0)            ;
                                ;
  fmul   dword ptr [_PolyDX]     ;
  fmul   dword ptr [_FUDGE_FACTOR];  Sshhh! don't tell anyone this is here :)
                                ;
  fmul   dword ptr [__tmp_fixone];
  fistp  dword ptr [_HorizontalDC];
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate horizontal texture delta values for the polygon if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF __INTERP_UV__             ;
                                ;
  fld   dword ptr [_FLTTextureX3];
  fsub  dword ptr [_FLTTextureX1]; (u3 - u1)
  fld   dword ptr [ebx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VB->Scr.y - VA->Scr.y)
  fmulp st(1),st(0)             ;
                                ;
  fld   dword ptr [ecx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VC->Scr.y - VA->Scr.y)
  fld   dword ptr [_FLTTextureX2];
  fsub  dword ptr [_FLTTextureX1]; (u2 - u1)
  fmulp st(1),st(0)             ;
                                ;
  fsubrp st(1),st(0)            ;
                                ;
  fmul   dword ptr [_PolyDX]     ;
  fmul   dword ptr [_FUDGE_FACTOR];  Sshhh! don't tell anyone this is here :)
                                ;
  fstp  dword ptr [_HorizontalDU]; Keep in floating point
                                ;
  fld   dword ptr [_FLTTextureY3];
  fsub  dword ptr [_FLTTextureY1]; (v3 - v1)
  fld   dword ptr [ebx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VB->Scr.y - VA->Scr.y)
  fmulp st(1),st(0)             ;
                                ;
  fld   dword ptr [ecx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VC->Scr.y - VA->Scr.y)
  fld   dword ptr [_FLTTextureY2];
  fsub  dword ptr [_FLTTextureY1]; (v2 - v1)
  fmulp st(1),st(0)             ;
                                ;
  fsubrp st(1),st(0)            ;
                                ;
  fmul   dword ptr [_PolyDX]     ;
  fmul   dword ptr [_FUDGE_FACTOR];  Sshhh! don't tell anyone this is here :)
                                ;
  fstp  dword ptr [_HorizontalDV]; Keep in floating point
                                ;
;----------------------------------------------------------------------------
; Calculate horizontal 1/z delta
;----------------------------------------------------------------------------
                                ;
  fld   dword ptr [_FLTZ3]       ;
  fsub  dword ptr [_FLTZ1]       ; (z3 - z1)
  fld   dword ptr [ebx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VB->Scr.y - VA->Scr.y)
  fmulp st(1),st(0)             ;
                                ;
  fld   dword ptr [ecx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VC->Scr.y - VA->Scr.y)
  fld   dword ptr [_FLTZ2]       ;
  fsub  dword ptr [_FLTZ1]       ; (z2 - z1)
  fmulp st(1),st(0)             ;
                                ;
  fsubrp st(1),st(0)            ;
                                ;
  fmul   dword ptr [_PolyDX]     ;
  fmul   dword ptr [_FUDGE_FACTOR];  Sshhh! don't tell anyone this is here :)
                                ;
  fstp  dword ptr [_HorizontalDZ]; Keep in floating point
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate horizontal env delta values for the polygon if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF __INTERP_ENV__            ;
                                ;
  fld   dword ptr [_FLTEnvX3]    ;
  fsub  dword ptr [_FLTEnvX1]    ; (s3 - s1)
  fld   dword ptr [ebx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VB->Scr.y - VA->Scr.y)
  fmulp st(1),st(0)             ;
                                ;
  fld   dword ptr [ecx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VC->Scr.y - VA->Scr.y)
  fld   dword ptr [_FLTEnvX2]    ;
  fsub  dword ptr [_FLTEnvX1]    ; (s2 - s1)
  fmulp st(1),st(0)             ;
                                ;
  fsubrp st(1),st(0)            ;
                                ;
  fmul   dword ptr [_PolyDX]     ;
  fmul   dword ptr [_FUDGE_FACTOR];  Sshhh! don't tell anyone this is here :)
                                ;
  fmul   dword ptr [__tmp_fixone];
  fistp  dword ptr [_HorizontalDS];
                                ;
  fld   dword ptr [_FLTEnvY3]    ;
  fsub  dword ptr [_FLTEnvY1]    ; (t3 - t1)
  fld   dword ptr [ebx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VB->Scr.y - VA->Scr.y)
  fmulp st(1),st(0)             ;
                                ;
  fld   dword ptr [ecx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VC->Scr.y - VA->Scr.y)
  fld   dword ptr [_FLTEnvY2]    ;
  fsub  dword ptr [_FLTEnvY1]    ; (t2 - t1)
  fmulp st(1),st(0)             ;
                                ;
  fsubrp st(1),st(0)            ;
                                ;
  fmul   dword ptr [_PolyDX]     ;
  fmul   dword ptr [_FUDGE_FACTOR];  Sshhh! don't tell anyone this is here :)
                                ;
  fmul   dword ptr [__tmp_fixone];
  fistp  dword ptr [_HorizontalDT];
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate OneOverTopDY
;----------------------------------------------------------------------------
                                ;
  Mov  edx,[_TopHeight]          ;
  Cmp  edx,0                    ;
  Jle @@SkipTopYHeightDelta     ;
                                ;
  fld  dword ptr [ebx].ScrY     ;
  fsub dword ptr [eax].ScrY     ;
                                ;
  fld1                          ;
  fdivrp  st(1),st(0)           ; 1.0 / (VB->Scr.y - VA->Scr.y)
  fstp  dword ptr [_OneOverTopDY];
                                ;
 @@SkipTopYHeightDelta:         ;
                                ;
;----------------------------------------------------------------------------
; Calculate OneOverBotDY
;----------------------------------------------------------------------------
                                ;
  Mov  edx,[_BottomHeight]       ;
  Cmp  edx,0                    ;
  Jle @@SkipBotYHeightDelta     ;

  fld  dword ptr [ecx].ScrY     ;
  fsub dword ptr [ebx].ScrY     ;
                                ;
  fld1                          ;
  fdivrp  st(1),st(0)           ; 1.0 / (VC->Scr.y - VB->Scr.y)
  fstp  dword ptr [_OneOverBotDY];
                                ;
 @@SkipBotYHeightDelta:         ;
                                ;
;----------------------------------------------------------------------------
; Calculate y-deltas. Done only when necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF __INTERP_EXTRA__          ;
                                ;
  fld  dword ptr [_PolyDX]       ;
  fldz                          ;
  fcompp                        ; if (plydx < 0.0)
                                ;
  Push  eax                     ; Save first vertex pointer
  fstsw ax                      ;
  sahf                          ;
                                ;
  Pop   eax                     ;
                                ;
  jae @@ShortSideLeftDelta      ;
                                ;
;----------------------------------------------------------------------------
; Calculate long gouraud deltas if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF __INTERP_C__              ;
                                ;
  fld  dword ptr [_FLTInten3]    ;
  fsub dword ptr [_FLTInten1]    ;
  fmul dword ptr [_OneOverTotalDY];  (c3 - c1) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDSide1DC]  ;  convert to fixed point
                                ;
  fild  dword ptr [_CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1dc
  fadd  dword ptr [_FLTInten1]   ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDSide1C]   ; convert to fixed point
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate long texture deltas if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF  __INTERP_UV__            ;
                                ;
  fld  dword ptr [_FLTTextureX3] ;
  fsub dword ptr [_FLTTextureX1] ;
  fmul dword ptr [_OneOverTotalDY];  (u3 - u1) * OneOverDY
                                ;
  fst  dword ptr [_FXDSide1DU]   ; keep as floating point
                                ;
  fild  dword ptr [_CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1du
  fadd  dword ptr [_FLTTextureX1];
  fstp  dword ptr [_FXDSide1U]   ; keep as floating point
                                ;
  fld  dword ptr [_FLTTextureY3] ;
  fsub dword ptr [_FLTTextureY1] ;
  fmul dword ptr [_OneOverTotalDY];  (v3 - v1) * OneOverDY
                                ;
  fst  dword ptr [_FXDSide1DV]   ; keep floating point
                                ;
  fild  dword ptr [_CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1dv
  fadd  dword ptr [_FLTTextureY1];
  fstp  dword ptr [_FXDSide1V]   ; keep floating point
                                ;
                                ;
  fld  dword ptr [_FLTZ3]        ;
  fsub dword ptr [_FLTZ1]        ;
  fmul dword ptr [_OneOverTotalDY];  (z3 - z1) * OneOverDY
                                ;
  fst  dword ptr [_FXDSide1DZ]   ; keep floating point
                                ;
  fild  dword ptr [_CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1dz
  fadd  dword ptr [_FLTZ1]       ;
  fstp  dword ptr [_FXDSide1Z]   ; keep floating point
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate long env deltas if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF  __INTERP_ENV__           ;
                                ;
  fld  dword ptr [_FLTEnvX3]     ;
  fsub dword ptr [_FLTEnvX1]     ;
  fmul dword ptr [_OneOverTotalDY];  (s3 - s1) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDSide1DS]  ;  convert to fixed point
                                ;
  fild  dword ptr [_CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1ds
  fadd  dword ptr [_FLTEnvX1]    ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDSide1S]   ; convert to fixed point
                                ;
  fld  dword ptr [_FLTEnvY3]     ;
  fsub dword ptr [_FLTEnvY1]     ;
  fmul dword ptr [_OneOverTotalDY];  (t3 - t1) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDSide1DT]  ;  convert to fixed point
                                ;
  fild  dword ptr [_CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1dt
  fadd  dword ptr [_FLTEnvY1]    ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDSide1T]   ; convert to fixed point
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Jump to the next section
;----------------------------------------------------------------------------
                                ;
  Jmp  @@BeginRendering         ; Jump to the actual rasterization process
 @@ShortSideLeftDelta:          ;
                                ;
;----------------------------------------------------------------------------
; Calculate top deltas if the top of the polygon exists
;----------------------------------------------------------------------------
                                ;
  Mov edx,[_TopHeight]           ;
  Cmp edx,0                     ;
  Jle @@CalcBottomHeightDeltas  ; check to make sure there is a top
                                ;
;----------------------------------------------------------------------------
; Calculate top short gouraud deltas if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF __INTERP_C__              ;
                                ;
  fld  dword ptr [_FLTInten2]    ;
  fsub dword ptr [_FLTInten1]    ;
  fmul dword ptr [_OneOverTopDY] ;  (c2 - c1) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDSide1DC]  ;  convert to fixed point
                                ;
  fild  dword ptr [_CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1dc
  fadd  dword ptr [_FLTInten1]   ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDSide1C]   ; convert to fixed point
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate short top texture deltas if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF  __INTERP_UV__            ;
                                ;
  fld  dword ptr [_FLTTextureX2] ;
  fsub dword ptr [_FLTTextureX1] ;
  fmul dword ptr [_OneOverTopDY] ;  (u2 - u1) * OneOverDY
                                ;
  fst   dword ptr [_FXDSide1DU]  ; keep floating point
                                ;
  fild  dword ptr [_CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1du
  fadd  dword ptr [_FLTTextureX1];
  fstp  dword ptr [_FXDSide1U]   ; keep floating point
                                ;
  fld  dword ptr [_FLTTextureY2] ;
  fsub dword ptr [_FLTTextureY1] ;
  fmul dword ptr [_OneOverTopDY] ;  (v2 - v1) * OneOverDY
                                ;
  fst   dword ptr [_FXDSide1DV]  ; keep floating point
                                ;
  fild  dword ptr [_CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1dv
  fadd  dword ptr [_FLTTextureY1];
  fstp  dword ptr [_FXDSide1V]   ; keep floating point
                                ;
  fld  dword ptr [_FLTZ2]        ;
  fsub dword ptr [_FLTZ1]        ;
  fmul dword ptr [_OneOverTopDY] ;  (z2 - z1) * OneOverDY
                                ;
  fst   dword ptr [_FXDSide1DZ]  ; keep floating point
                                ;
  fild  dword ptr [_CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1dz
  fadd  dword ptr [_FLTZ1]       ;
  fstp  dword ptr [_FXDSide1Z]   ; keep floating point
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate long env deltas if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF  __INTERP_ENV__           ;
                                ;
  fld  dword ptr [_FLTEnvX2]     ;
  fsub dword ptr [_FLTEnvX1]     ;
  fmul dword ptr [_OneOverTopDY] ;  (s2 - s1) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDSide1DS]  ;  convert to fixed point
                                ;
  fild  dword ptr [_CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1ds
  fadd  dword ptr [_FLTEnvX1]    ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDSide1S]   ; convert to fixed point
                                ;
  fld  dword ptr [_FLTEnvY2]     ;
  fsub dword ptr [_FLTEnvY1]     ;
  fmul dword ptr [_OneOverTopDY] ;  (t3 - t1) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDSide1DT]  ;  convert to fixed point
                                ;
  fild  dword ptr [_CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1dt
  fadd  dword ptr [_FLTEnvY1]    ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDSide1T]   ; convert to fixed point
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Check to see if the bottom half of the polygon exists
;----------------------------------------------------------------------------
                                ;
 @@CalcBottomHeightDeltas:      ;
                                ;
  Mov  edx,[_BottomHeight]      ;
  Cmp  edx,0                    ;
  Jle  @@BeginRendering         ;
                                ;
;----------------------------------------------------------------------------
; Calculate top short gouraud deltas if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF __INTERP_C__              ;
                                ;
  fld  dword ptr [_FLTInten3]    ;
  fsub dword ptr [_FLTInten2]    ;
  fmul dword ptr [_OneOverBotDY] ;  (c3 - c2) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDSide2DC]  ;  convert to fixed point
                                ;
  fild  dword ptr [_CeilY2]      ;
  fsub  dword ptr [ebx].ScrY    ; (CeilY2 - VB->Scr.y)
  fmulp st(1),st(0)             ; (CeilY2 - VB->Scr.y) * fltside2dc
  fadd  dword ptr [_FLTInten2]   ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDSide2C]   ; convert to fixed point
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate short top texture deltas if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF  __INTERP_UV__            ;
                                ;
  fld  dword ptr [_FLTTextureX3] ;
  fsub dword ptr [_FLTTextureX2] ;
  fmul dword ptr [_OneOverBotDY] ; (u3 - u2) * OneOverDY
                                ;
  fst  dword ptr [_FXDSide2DU]   ; keep floating point
                                ;
  fild  dword ptr [_CeilY2]      ;
  fsub  dword ptr [ebx].ScrY    ; (CeilY2 - VB->Scr.y)
  fmulp st(1),st(0)             ; (CeilY2 - VB->Scr.y) * fltside2du
  fadd  dword ptr [_FLTTextureX2];
  fstp  dword ptr [_FXDSide2U]   ; convert to fixed point
                                ;
  fld  dword ptr [_FLTTextureY3] ;
  fsub dword ptr [_FLTTextureY2] ;
  fmul dword ptr [_OneOverBotDY] ;  (v2 - v1) * OneOverDY
                                ;
  fst  dword ptr [_FXDSide2DV]   ;  keep floating point
                                ;
  fild  dword ptr [_CeilY2]      ;
  fsub  dword ptr [ebx].ScrY    ; (CeilY2 - VB->Scr.y)
  fmulp st(1),st(0)             ; (CeilY2 - VB->Scr.y) * fltside2dv
  fadd  dword ptr [_FLTTextureY2];
  fstp  dword ptr [_FXDSide2V]   ; keep floating point
                                ;
  fld  dword ptr [_FLTZ3]        ;
  fsub dword ptr [_FLTZ2]        ;
  fmul dword ptr [_OneOverBotDY] ;  (z2 - z1) * OneOverDY
                                ;
  fst  dword ptr [_FXDSide2DZ]   ;  keep floating point
                                ;
  fild  dword ptr [_CeilY2]      ;
  fsub  dword ptr [ebx].ScrY    ; (CeilY2 - VB->Scr.y)
  fmulp st(1),st(0)             ; (CeilY2 - VB->Scr.y) * fltside2dv
  fadd  dword ptr [_FLTZ2]       ;
  fstp  dword ptr [_FXDSide2Z]   ; keep floating point
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate long env deltas if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF  __INTERP_ENV__           ;
                                ;
  fld  dword ptr [_FLTEnvX3]     ;
  fsub dword ptr [_FLTEnvX2]     ;
  fmul dword ptr [_OneOverBotDY] ;  (s3 - s2) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDSide2DS]  ;  convert to fixed point
                                ;
  fild  dword ptr [_CeilY2]      ;
  fsub  dword ptr [ebx].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1ds
  fadd  dword ptr [_FLTEnvX2]    ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDSide2S]   ; convert to fixed point
                                ;
  fld  dword ptr [_FLTEnvY3]     ;
  fsub dword ptr [_FLTEnvY2]     ;
  fmul dword ptr [_OneOverBotDY] ;  (t3 - t2) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDSide2DT]  ;  convert to fixed point
                                ;
  fild  dword ptr [_CeilY2]      ;
  fsub  dword ptr [ebx].ScrY    ; (CeilY2 - VB->Scr.y)
  fmulp st(1),st(0)             ; (CeilY2 - VB->Scr.y) * fltside2dt
  fadd  dword ptr [_FLTEnvY2]    ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDSide2T]   ; convert to fixed point
                                ;
ENDIF                           ;
                                ;
ENDIF                           ;  __INTERP_EXTRA__
;----------------------------------------------------------------------------
; Lets do some rendering!!!
;----------------------------------------------------------------------------
                                ;
 @@BeginRendering:              ; the start of the rasterization process
                                ;
  Push  eax                     ; Save vertex A pointer
                                ;
  Mov   eax,[_CeilY1]           ;
  Mov   edx,_LineSize         ;
  IMul  edx                     ;
  Add   eax,_VirtualAddress   ; add in the address of the virtual page
  Mov   [_YAddress],eax         ;
                                ;
  Pop   eax                     ; restore vertex A pointer
                                ;
;----------------------------------------------------------------------------
; See if we need to do the top half of the triangle
;----------------------------------------------------------------------------
                                ;
  Mov  edx,[_TopHeight]          ;
  Cmp  edx,0                    ;
  Jle  @@SkipTopTriangleRender  ;
                                ;
;----------------------------------------------------------------------------
; Calculate the short side data
;----------------------------------------------------------------------------
                                ;
  fld  dword ptr [ebx].ScrX     ;
  fsub dword ptr [eax].ScrX     ; (VB->Scr.x - VA->Scr.x)
  fmul dword ptr [_OneOverTopDY] ; (VB->Scr.x - VA->Scr.x) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDShortDX]  ;  store fixed point
                                ;
  fild  dword ptr [_CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ;
  fmulp st(1),st(0)             ;
  fadd  dword ptr [eax].ScrX    ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDShortX]   ;  store fixed point
                                ;
;----------------------------------------------------------------------------
; Compare deltas to see which order we draw the polygon
;----------------------------------------------------------------------------
                                ;
  Mov  edx,[_FXDLongDX]          ;
  Cmp  edx,[_FXDShortDX]         ; if (longdx < shortdx)
                                ;
  Jge @@LongSideRight           ;
                                ;
  INNER_LOOP _FXDLong, _FXDShort, _TopHeight, _FXDSide1
                                ;
  Jmp @@SkipTopTriangleRender   ;
 @@LongSideRight:               ;
                                ;
  INNER_LOOP _FXDShort, _FXDLong, _TopHeight, _FXDSide1
                                ;
 @@SkipTopTriangleRender:       ;
                                ;
;----------------------------------------------------------------------------
; See if we need to do the bottom half of the triangle
;----------------------------------------------------------------------------
                                ;
  Mov  edx,[_BottomHeight]       ;
  Cmp  edx,0                    ;
  Jle  @@SkipBottomTriangleRender;
                                ;
  Mov  eax,[_VertPtrA]           ;  re-grab the vertex pointers
  Mov  ebx,[_VertPtrB]           ;
  Mov  ecx,[_VertPtrC]           ;
                                ;
;----------------------------------------------------------------------------
; Calculate the short side data for bottom
;----------------------------------------------------------------------------
                                ;
  fld  dword ptr [ecx].ScrX     ;
  fsub dword ptr [ebx].ScrX     ; (VC->Scr.x - VB->Scr.x)
  fmul dword ptr [_OneOverBotDY] ; (VC->Scr.x - VB->Scr.x) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDShortDX]  ;  store fixed point
                                ;
  fild  dword ptr [_CeilY2]      ;
  fsub  dword ptr [ebx].ScrY    ;
  fmulp st(1),st(0)             ;
  fadd  dword ptr [ebx].ScrX    ;
  fmul  dword ptr [__tmp_fixone] ;
  fistp dword ptr [_FXDShortX]   ;  store fixed point
                                ;
;----------------------------------------------------------------------------
; Compare deltas to see which order we draw the polygon for bottom
;----------------------------------------------------------------------------
                                ;
  Mov  edx,[_FXDLongDX]          ;
  Cmp  edx,[_FXDShortDX]         ; if (longdx < shortdx)
                                ;
  Jge @@RasterLongSideRight     ;
                                ;
  INNER_LOOP _FXDShort, _FXDLong, _BottomHeight, _FXDSide2
                                ;
  Jmp @@ExitPolygonFiller       ;
 @@RasterLongSideRight:         ;
                                ;
  INNER_LOOP _FXDLong, _FXDShort, _BottomHeight, _FXDSide1
                                ;
 @@SkipBottomTriangleRender:    ;
 @@ExitPolygonFiller:           ;
