;============================================================================
;                    Inertia Realtime 3D Rendering Engine
;         Copyright (c) 1996,1997 Alex Chalfin. All Rights Reserved.
;          DISTRIBUTION OF THIS SOURCE CODE IS STRICTLY PROHIBITED
;============================================================================

;===================
; PLYPHPT.ASM
; code: Alex Chalfin
;===================

; Phong Texture (perspective) polygon filler
;
;================================================================
; Polygon specific defines
;================================================================
;
__INTERP_EXTRA__  EQU 0        ; must be defined for everything but flat
;__INTERP_C__      EQU 0        ; must be defined for goruaud shading
__INTERP_UV__     EQU 0        ; must be defined for texture mapping
__INTERP_ENV__    EQU 0        ; must be defined for env mapping (phong)
;

.386p                           ; 486 protected mode
.387
.Model Flat,C                   ; with flat memory model

DATA SEGMENT PARA PUBLIC USE32  ;
                                ;
public _PhongPerspTexturePolygonASM;
                                ;
include strucdef.asm            ; Inertia structure references
include varsasm.inc             ; external variable references
                                ;
DATA ENDS                       ;
                                ;
;================================================================
; Code Segment
;================================================================
                                ;
CODE SEGMENT PARA PUBLIC USE32  ;
                                ;
;===========================================================================
; Macros
;===========================================================================
                                ;
;================================================================
;  AFFINE_LOOP
;================================================================
                                ; on entry, edi points to the correct
                                ; screen location to start drawing.
AFFINE_LOOP Macro               ;
Local @@FillLoop                ;
  Mov  eax,[_LAffU]              ;
  Mov  ebx,[_LAffV]              ;
                                ;
  Sar  eax,8                    ;
  Sar  ebx,8                    ;
                                ;
  Mov  [_UStart],eax             ; save U start
  Mov  [_VStart],ebx             ; save V start 
                                ;
  Mov  edx,[_AffDU]              ;
  Mov  esi,[_AffDV]              ;
                                ;
  Sar  edx,8                    ;
  Sar  esi,8                    ;
                                ;
  Mov  [_HDU8],edx               ;
  Mov  [_HDV8],esi               ;
                                ;
  Mov  eax,[_SStart]             ;
  Mov  ebx,[_TStart]             ;
                                ;
  Sar  eax,8                    ;
  Sar  ebx,8                    ;
                                ;
  Mov  [_SStart8],eax            ; save SStart
  Mov  [_TStart8],ebx            ; save TStart
                                ;
  Mov  edx,[_HorizontalDS]       ;
  Mov  esi,[_HorizontalDT]       ;
                                ;
  Sar  edx,8                    ;
  Sar  esi,8                    ;
                                ;
  Mov  [_HDS8],edx               ;
  Mov  [_HDT8],esi               ;
                                ;
  Mov  esi,[_TStart8]            ; retreive TStart
  Mov  ecx,[_SStart8]            ; retrieve SStart
  Mov  ebx,[_VStart]             ; retrieve VStart
  Mov  eax,[_UStart]             ; retrieve UStart
                                ;
                                ; add in the intensity offset
                                ; On entering loop:
                                ; edi = screen position
                                ; esi = 8.8 env T value
                                ; ecx = 8.8 env S value
                                ; eax = 8.8 texture U value
                                ; ebx = 8.8 texture V value
  Push  ebp                     ;
  Mov  ebp,[_TextureOfsTable]    ;
 @@FillLoop:                    ;
                                ;
  Mov  ecx,[_SStart8]            ;
                                ;
  Mov  edx,esi                  ;
  Mov  dl,ch                    ;
  Add  ecx,[_HDS8]               ;
  And  edx,0ffffh               ;
                                ;
  Mov  ebp,[_IBufferTable]       ;
                                ;
  Add  esi,[_HDT8]               ;
  Mov  [_SStart8],ecx            ;
                                ;
  Mov  ch,[ebp+edx]             ; get phong map color
                                ;
  Mov  ebp,[_TextureOfsTable]    ;
                                ;
  Mov  dh,bh                    ;
  Mov  dl,ah                    ;
                                ;
  Add  eax,[_HDU8]               ;
  Add  ebx,[_HDV8]               ;
                                ;
  And  ecx,0ffffh               ;
                                ;
  Mov  cl,[ebp+edx]             ; texture color
  Mov  ebp,[_IntensityTable]     ;
                                ;
  Inc  edi                      ;
                                ;
  Mov  dl,[ebp+ecx]             ; get shade color
                                ;
  Xor  ecx,ecx                  ;
  Dec  [_DrawWidth]              ;
                                ;
  Mov  [edi-1],dl               ;
                                ;
  Jnz @@FillLoop                ;
  Pop  ebp                      ;
                                ;
  Mov  eax,[_SStart]             ;
  Mov  edx,[_TStart]             ;
                                ;
  Mov  ecx,[_HorizontalDS]       ;
  Mov  esi,[_HorizontalDT]       ;
                                ;
  Shl  ecx,4                    ;
  Shl  esi,4                    ;
                                ;
  Add  eax,ecx                  ;
  Add  edx,esi                  ;
                                ;
  Mov  [_SStart],eax             ;
  Mov  [_TStart],edx             ;
                                ;
EndM                            ;
                                ;
;================================================================
;  PERSP_LOOP
;================================================================
                                ; on entry, edi points to the correct
                                ; screen location to start drawing.
                                ; No registers need to be saved.
                                ; the FPU stack should be clear after execution
PERSP_LOOP Macro                ;
Local @@SubdivisionWhile,@@ExitSubWhile

 @@SubdivisionWhile:            ;
  Mov  eax,[_SpanCount]          ;
  Cmp  eax,0                    ;
  Jle  @@ExitSubWhile           ;
                                ;
  Mov  ebx,eax                  ;
  Sub  eax,PERSPSPAN            ;
                                ;
  Mov  [_SpanCount],eax          ;
                                ;
;----------------------------------------------------------------------
; This piece of code is MIN(SpanCount, PERSPSPAN) without branching.
; Branching can have a *huge* penalty on ppro machines.
;----------------------------------------------------------------------
                                ;
  Mov  eax,ebx                  ; copy the current span width
  Mov  ebx,PERSPSPAN            ; the length of a affine span
  Cmp  ebx,eax                  ; compare the two
  Sbb  ecx,ecx                  ; if (ebx < eax), ecx = 0xffffffff, else ecx = 0
  Mov  edx,ecx                  ; copy the flag over
  Not  ecx                      ; take the bitwise inverse
  And  eax,ecx                  ; mask the origional value
  And  ebx,edx                  ; mask the max value
  Or   eax,ebx                  ; store the value
                                ;
  Mov  [_DrawWidth],eax          ;
                                ;
;-----------------------------------------------------------------------
                                ;
  fild  dword ptr [_DrawWidth]   ;
  fld   st(0)                   ;
                                ;
  fmul  dword ptr [_HorizontalDU];
  fadd  dword ptr [_hpu]         ;
  fstp  dword ptr [_hpu]         ;
                                ;
  fld   st(0)                   ;
                                ;
  fmul  dword ptr [_HorizontalDV];
  fadd  dword ptr [_hpv]         ;
  fstp  dword ptr [_hpv]         ;
                                ;
  fmul  dword ptr [_HorizontalDZ];
  fadd  dword ptr [_hpz]         ;
  fst   dword ptr [_hpz]         ; store without pop to keep hpz value on the
                                ; FPU stack
;---------------------------------------------------------------------------
                                ;
  fld1                          ; load 1.0
  fdivrp st(1),st(0)                  ; 1.0 / hpz
                                ;
  fld    st(0)                  ; store 1.0/hpz again
  fmul   dword ptr [_hpu]        ;
  fmul   dword ptr [__tmp_fixone];
                                ;
  fist   dword ptr [_RAffU]      ; store RAffU
  fisub  dword ptr [_LAffU]      ; calculate (RAffU - LAffU) in 16.16 fixed
                                ;
  fmul   dword ptr [_DivTable+eax*4];  eax = DrawWidth
                                ;
  fistp  dword ptr [_AffDU]      ; affine delta U in 16.16 fixed point
                                ;
                                ;
  fmul   dword ptr [_hpv]        ;
  fmul   dword ptr [__tmp_fixone];
                                ;
  fist   dword ptr [_RAffV]      ; store RAffU
  fisub  dword ptr [_LAffV]      ; calculate (RAffU - LAffU) in 16.16 fixed
                                ;
  fmul   dword ptr [_DivTable+eax*4];  eax = DrawWidth
                                ;
  fistp  dword ptr [_AffDV]      ; affine delta U in 16.16 fixed point
                                ;
;-----------------------------------------------------------------------
; Draw the affine pixels. LAffU,LAffV are the starting texel
; values in 16.16 fixed and AffDU,AffDV are the deltas
;-----------------------------------------------------------------------
                                ;
  AFFINE_LOOP                   ;
                                ;
;-----------------------------------------------------------------------
                                ;
  Mov    eax,[_RAffU]            ;
  Mov    ebx,[_RAffV]            ;
                                ;
  Mov    [_LAffU],eax            ;
  Mov    [_LAffV],ebx            ;
                                ;
  Jmp  @@SubdivisionWhile       ;
                                ;
 @@ExitSubWhile:                ;
                                ;
EndM                            ;
                                ;
;================================================================
;  INNER_LOOP
;================================================================
                                ; All registers are available for loop
                                ; macro. Nothing needs to be saved/restored
                                ; it is all handled in PERSPASM.INC
                                ;
INNER_LOOP Macro  Left, Right, Height, Side;
Local  @@HeightLooper,@@ExitHeight ; Local labels
                                ;
 @@HeightLooper:                ;
  Mov  eax,[Height]             ;
  Cmp  eax,0                    ;
  Jle @@ExitHeight              ;
                                ;
  Dec  [Height]                 ; decrement the height value by 1
                                ;
  Mov  ebx,[Left&X]             ;
  Add  ebx,65535                ;
  And  ebx,0ffff0000h           ;
                                ;
  Mov  edx,ebx                  ;
  Sub  edx,[Left&X]             ; edx = (ceil(LeftX) - LeftX) in 16.16 fixed
                                ;
  Mov  [__tmp_flt],edx           ;
                                ;
  fild dword ptr [__tmp_flt]     ;
  fmul dword ptr [__tmp_fxd2flt] ; convert the fixed value to a float
                                ;
  fld  st(0)                    ; load the value again
;--------------------------------
; Sub-pixel correction
;--------------------------------

  Mov  ecx,edx                  ;
                                ;
  Mov  eax,[_HorizontalDS]       ;
  IMul edx                      ; mul (ceil(LeftX) - LeftX) by HorizontalDS
  Shrd eax,edx,16               ; fix for sub-pixel adjustment
                                ;
  Add  eax,[Side&S]             ; eax is the starting horizontal color
  Mov  [_SStart],eax             ;
                                ;
  Mov  eax,[_HorizontalDT]       ;
  IMul ecx                      ; mul (ceil(LeftX) - LeftX) by HorizontalDT
  Shrd eax,edx,16               ; fix for sub-pixel adjustment
                                ;
  Add  eax,[Side&T]             ; eax is the starting horizontal color
  Mov  [_TStart],eax             ;
                                ;
                                ;
  fmul dword ptr [_HorizontalDU] ; perform sub-pixel correction
  fadd dword ptr [Side&U]       ; add in starting U/Z value
                                ;
  fstp dword ptr [_hpu]          ; store the horizontal perspective u value
                                ;-------------------------------------------
  fld  st(0)                    ;
  fmul dword ptr [_HorizontalDV] ; perform sub-pixel correction
  fadd dword ptr [Side&V]       ; add in starting V/Z value
                                ;
  fstp dword ptr [_hpv]          ; store the horizontal perspective v value
                                ;-------------------------------------------
  fmul dword ptr [_HorizontalDZ] ; perform sub-pixel correction
  fadd dword ptr [Side&Z]       ; add in starting 1/Z value
                                ;
  fst  dword ptr [_hpz]          ; store the horizontal perspective z value
;-------------------------------------------------------------------------
                                ;
  fld1                          ; load 1.0
  fdivrp st(1),st(0)                  ; calculate 1/hpz for the starting loc
                                ;
  fld    st(0)                  ; reload the 1/z value
                                ;
  fmul   dword ptr [_hpu]        ; get the perspective correct U coordiante.
  fmul   dword ptr [__tmp_fixone]; convert to fixed point
                                ;
  fistp  dword ptr [_LAffU]      ; store the left affine U value
                                ;
  fmul   dword ptr [_hpv]        ; get the perspective correct V coordiante.
  fmul   dword ptr [__tmp_fixone]; convert to fixed point
                                ;
  fistp  dword ptr [_LAffV]      ; store the left affine V value
;--------------------------------------------------------------------------
                                ;
  Mov    eax,[Left&X]           ;
  Mov    ebx,[Right&X]          ;
                                ;
  Add    eax,65535              ;
  Dec    ebx                    ;
                                ;
  Sar    eax,16                 ;
  Sar    ebx,16                 ;
                                ;
  Sub    ebx,eax                ;
  Inc    ebx                    ;
                                ;
  Mov    edi,eax                ;
  Add    edi,[_YAddress]         ; edi points to the screen location
                                ;
  Mov    [_SpanCount],ebx        ; the length of the span
;--------------------------------------------------------------------------
                                ;
  PERSP_LOOP                    ; perspective span loop
                                ;
;--------------------------------------------------------------------------
                                ;
  Mov  edx,[_YAddress]           ; go to the next scanline
  Add  edx,_LineSize           ;
  Mov  [_YAddress],edx           ; YAddress += LineSize
                                ;
  Mov  eax,[Side&S]             ; increment the S span
  Add  eax,[Side&DS]            ;
  Mov  [Side&S],eax             ;
                                ;
  Mov  eax,[Side&T]             ; increment the T span
  Add  eax,[Side&DT]            ;
  Mov  [Side&T],eax             ;
                                ;
  fld  dword ptr [Side&U]       ; increment the U span
  fadd dword ptr [Side&DU]      ;
  fstp dword ptr [Side&U]       ;
                                ;
  fld  dword ptr [Side&V]       ; increment the V span
  fadd dword ptr [Side&DV]      ;
  fstp dword ptr [Side&V]       ;
                                ;
  fld  dword ptr [Side&Z]       ; increment the Z span
  fadd dword ptr [Side&DZ]      ;
  fstp dword ptr [Side&Z]       ;
                                ;
  Mov  eax,[Left&X]             ;
  Mov  ebx,[Right&X]            ;
  Add  eax,[Left&DX]            ;
  Add  ebx,[Right&DX]           ;
  Mov  [Left&X],eax             ; LeftX += LeftDX
  Mov  [Right&X],ebx            ; RightX += RightDX
                                ;
  Jmp @@HeightLooper            ;
                                ;
 @@ExitHeight:                  ;
EndM                            ;
                                ;
;===========================================================================
; FlatPerspTexturePolygonASM
;===========================================================================
_PhongPerspTexturePolygonASM Proc Uses edi eax,_Triangle:DWord, _VirtualAddress:DWord, _LineSize:DWord
                                ;
  Pushad

  Mov  edi,_Triangle           ;
                                ;
  Mov   eax,[edi].Texture       ;
  Mov   [_TextureOfsTable],eax   ;  grab and store the texture pointer
                                ;
  Mov   eax,[edi].IBuffer       ;
  Mov   [_IBufferTable],eax      ;  grab and store the texture pointer
                                ;
include perspasm.inc            ; include perspective interpolation
                                ;
  Popad

  Ret                           ;
                                ;
_PhongPerspTexturePolygonASM EndP  ;


CODE ENDS
END
