;
;                    Inertia Realtime 3D Rendering Engine
;    Copyright (c) 1996, Alex Chalfin, Jeroen Bouwens. All Rights Reserved.
;          DISTRIBUTION OF THIS SOURCE CODE IS STRICTLY PROHIBITED
;

;
; PLYPHRET.ASM
; code: Alex Chalfin
;

; Transparent Phong Reflection polygon filler
;
;================================================================
; Polygon specific defines
;================================================================
;
__INTERP_EXTRA__  EQU 0        ; must be defined for everything but flat
;__INTERP_C__      EQU 0        ; must be defined for goruaud shading
;__INTERP_UV__     EQU 0        ; must be defined for texture mapping
__INTERP_ENV__    EQU 0        ; must be defined for env mapping (phong)
;

.386p                           ; 486 protected mode
.387
.Model Flat,C                   ; with flat memory model

DATA SEGMENT PARA PUBLIC USE32  ;
                                ;
public _TransPhongReflectionPolygonASM;
                                ;
include strucdef.asm            ;  Inertia structure references
include varsasm.inc             ; external variable references
                                ;
DATA ENDS                       ;
                                ;
;================================================================
; Code Segment
;================================================================
                                ;
CODE SEGMENT PARA PUBLIC USE32  ;
                                ;
;===========================================================================
; Macros
;===========================================================================
                                ;
                                ; All registers are available for loop
                                ; macro. Nothing needs to be saved/restored
                                ; it is all handled in AFFASM.INC
                                ;
INNER_LOOP Macro  Left, Right, Height, Side;
Local  @@HeightLooper,@@ExitHeight,@@SkipFill,@@FillLoop ; Local labels
                                ;
 @@HeightLooper:                ;
  Mov  eax,[Height]             ;
  Cmp  eax,0                    ;
  Jle @@ExitHeight              ;
                                ;
  Dec  [Height]                 ;  decrement the height value by 1
                                ;
  Mov  ebx,[Left&X]             ;
  Add  ebx,65535                ;
  And  ebx,0ffff0000h           ;
                                ;
  Mov  edx,ebx                  ;
  Sub  edx,[Left&X]             ;
                                ; edx contains sub-pixel correction factor
  Mov  esi,edx                  ;
                                ;
  Mov  eax,[_HorizontalDS]       ;
  IMul edx                      ;
  Shrd eax,edx,16               ; fix for sub-pixel adjustment
                                ;
  Add  eax,[Side&S]             ; eax is the starting horizontal color
                                ;
  Mov  [_SStart],eax             ;
                                ;
  Mov  edx,esi                  ; get sup-pixel correction factor
  Mov  eax,[_HorizontalDT]       ;
  IMul edx                      ;
  Shrd eax,edx,16               ; fix for sub-pixel adjustment
                                ;
  Add  eax,[Side&T]             ; eax is the starting horizontal color
                                ;
  Mov  [_TStart],eax             ;
                                ;
  Sar  ebx,16                   ;
                                ;
  Mov  ecx,[Right&X]            ;
  Dec  ecx                      ;
  Sar  ecx,16                   ;
                                ;
  Sub  ecx,ebx                  ;
  Inc  ecx                      ;
  Cmp  ecx,0                    ;
  Jle @@SkipFill                ;
                                ;
  Mov  [_Count],ecx              ;
                                ;
  Mov  edi,[_YAddress]           ;
  Add  edi,ebx                  ;
                                ;
  Mov  eax,[_SStart]             ;
  Mov  ebx,[_TStart]             ;
                                ;
  Sar  eax,8                    ;
  Sar  ebx,8                    ;
                                ;
  Mov  [_SStart],eax             ;
  Mov  [_TStart],ebx             ;
                                ;
  Mov  edx,[_HorizontalDS]       ;
  Mov  esi,[_HorizontalDT]       ;
                                ;
  Sar  edx,8                    ;
  Sar  esi,8                    ;
                                ;
  Mov  [_HDS8],edx               ;
  Mov  [_HDT8],esi               ;
                                ;
  Mov  eax,[_TStart]             ;
  Mov  ebx,[_SStart]             ;
                                ; add in the intensity offset
                                ; On entering loop:
                                ; edi = screen position
                                ; eax = 8.8 env T value
                                ; ebx = 8.8 env S value
  Xor   edx,edx                 ;
  Push  ebp                     ;
  Mov  ebp,[_TextureOfsTable]    ;
 @@FillLoop:                    ;
                                ;
  Mov  dl,bh                    ;
  Mov  dh,ah                    ;
                                ;
  Add  ebx,[_HDS8]               ;
  Add  eax,[_HDT8]               ;
                                ;
  Mov  ebp,[_IBufferTable]       ;
                                ;
  Mov  ch,[ebp+edx]             ; get phong map color
                                ;
  Mov  ebp,[_TextureOfsTable]    ;
                                ;
  Mov  cl,[ebp+edx]             ; texture color
  Mov  ebp,[_IntensityTable]     ;
                                ;
  Mov  dl,[ebp+ecx]             ; get shade color
  Mov  dh,[edi]                 ; get screen color
                                ;
  Mov  ebp,[_TransparencyTable]  ;
  Inc  edi                      ;
                                ;
  Mov  cl,[ebp+edx]             ;
                                ;
  Mov  [edi-1],cl               ;
                                ;
  Xor  ecx,ecx                  ;
  Dec  [_Count]                  ;
                                ;
  Jnz @@FillLoop                ;
  Pop  ebp                      ;
                                ;
 @@SkipFill:                    ;
                                ;
  Mov  edx,[_YAddress]           ; go to the next scanline
  Add  edx,_LineSize           ;
  Mov  [_YAddress],edx           ; YAddress += LineSize
                                ;
  Mov  eax,[Side&S]             ; increment the U span
  Add  eax,[Side&DS]            ;
  Mov  [Side&S],eax             ;
                                ;
  Mov  eax,[Side&T]             ; increment the U span
  Add  eax,[Side&DT]            ;
  Mov  [Side&T],eax             ;
                                ;
  Mov  eax,[Left&X]             ;
  Mov  ebx,[Right&X]            ;
  Add  eax,[Left&DX]            ;
  Add  ebx,[Right&DX]           ;
  Mov  [Left&X],eax             ; LeftX += LeftDX
  Mov  [Right&X],ebx            ; RightX += RightDX
                                ;
  Jmp @@HeightLooper            ;
                                ;
 @@ExitHeight:                  ;
EndM                            ;
                                ;
;===========================================================================
; CullPolys
;===========================================================================
_TransPhongReflectionPolygonASM Proc Uses edi eax,_Triangle:DWord, _VirtualAddress:DWord, _LineSize:DWord
                                ;
  Pushad

  Mov edi,_Triangle            ;
                                ;
  Mov   eax,[edi].Texture       ;
  Mov   [_TextureOfsTable],eax   ;  grab and store the texture pointer
                                ;
  Mov   eax,[edi].IBuffer       ;
  Mov   [_IBufferTable],eax      ;  grab and store the texture pointer
                                ;
include affasm.inc              ; include affine interpolation
                                ;
  Popad

  Ret                           ;
                                ;
_TransPhongReflectionPolygonASM EndP  ;


CODE ENDS
END
